from typing import (Callable,
                    Dict,
                    Union)

from pkg_resources import iter_entry_points

from . import plugins
from .models import (Resource,
                     URL)

_stop = object()

_DeserializersKey = Union[str, type(_stop)]
_Deserializer = Callable[[str], Resource]
_Deserializers = Dict[_DeserializersKey, _Deserializer]
_Deserializers = Dict[_DeserializersKey, _Deserializers]


def _load_deserializers() -> _Deserializers:
    registry = {entry_point.name: entry_point.load()
                for entry_point in iter_entry_points(plugins.__name__)}
    result = {}
    for name, deserializer in registry.items():
        id_ = plugins.deserialize_id(name)
        location = result
        for key in id_:
            location = location.setdefault(key, {})
        location[_stop] = deserializer
    return result


_deserializers = _load_deserializers()


def deserialize(string: str) -> Resource:
    result = URL.from_string(string)
    location = _deserializers
    for field in result._tuple:
        try:
            location = location[field]
        except KeyError:
            break
    try:
        deserializer = location[_stop]
    except KeyError as error:
        raise ValueError('Resource "{resource}" is not supported.'
                         .format(resource=string)) from error
    else:
        return deserializer(string)



from abc import (ABC,
                 abstractmethod)
from pathlib import PurePosixPath
from typing import (Any,
                    Generic,
                    Iterable,
                    Iterator,
                    Optional,
                    Type,
                    TypeVar,
                    Union)
from urllib.parse import (quote,
                          quote_plus,
                          unquote,
                          unquote_plus,
                          urlparse,
                          urlunparse)

from memoir import cached
from reprit.base import generate_repr

from .hints import Domain


# not using `namedtuple` to be weak referencable
class URL:
    _fields = ('scheme', 'authority', 'path_string',
               'params', 'query', 'fragment')
    __slots__ = ('_tuple', '__weakref__')

    # added for auto-completion support,
    # each of them will be overwritten afterwards in cycle
    # and made a `property`
    scheme = ''
    authority = ''
    path_string = ''
    params = ''
    query = ''
    fragment = ''

    for index, field in enumerate(_fields):
        locals()[field] = property(
                lambda self, index=index: self._tuple[index])
    del index, field

    def __init__(self,
                 scheme: str,
                 authority: str,
                 path_string: str,
                 params: str,
                 query: str,
                 fragment: str) -> None:
        self._tuple = (scheme, authority, path_string, params, query, fragment)

    def __hash__(self) -> int:
        return hash(self._tuple)

    def __eq__(self, other: 'URL') -> bool:
        if not isinstance(other, URL):
            return NotImplemented
        return self._tuple == other._tuple

    def __str__(self) -> str:
        return urlunparse(self._replace(path_string=quote(self.path_string,
                                                          '/%'),
                                        query=quote_plus(self.query, ':&='))
                          ._tuple)

    __repr__ = generate_repr(__init__)

    def _replace(self, **kwargs: str) -> 'URL':
        return type(self)(*map(kwargs.pop, self._fields, self._tuple))

    def join(self, part: str, *parts: str) -> 'URL':
        return self._replace(path_string=str(self.path.joinpath(part, *parts)),
                             params='',
                             query='',
                             fragment='')

    @cached.property_
    def path(self) -> PurePosixPath:
        return PurePosixPath(self.path_string)

    @classmethod
    def from_string(cls: Type[Domain], string: str) -> Domain:
        raw = urlparse(string)
        return cls(scheme=raw.scheme,
                   authority=raw.netloc,
                   path_string=unquote(raw.path),
                   params=raw.params,
                   query=unquote_plus(raw.query),
                   fragment=raw.fragment)


class Base(ABC):
    @property
    @abstractmethod
    def url(self) -> URL:
        pass

    @abstractmethod
    def exists(self) -> bool:
        pass

    @abstractmethod
    def __repr__(self) -> str:
        pass

    @abstractmethod
    def __str__(self) -> str:
        pass

    @abstractmethod
    def __hash__(self) -> int:
        pass

    @abstractmethod
    def __eq__(self, other: 'Base') -> bool:
        pass

    @classmethod
    @abstractmethod
    def from_string(cls: Type[Domain], string: str) -> Domain:
        pass


class Container(Base):
    @abstractmethod
    def join(self, part: str, *parts: str) -> Base:
        pass

    def __truediv__(self, other: str) -> Base:
        if not isinstance(other, str):
            return NotImplemented
        return self.join(other)

    @abstractmethod
    def __iter__(self) -> Iterator[Base]:
        pass


Element = TypeVar('Element')


class Stream(Base, Generic[Element]):
    @abstractmethod
    def open(self, **kwargs) -> Iterable[Element]:
        pass

    @abstractmethod
    def send(self, destination: Base, **kwargs: Any) -> None:
        pass

    @abstractmethod
    def receive(self, source: Base, **kwargs: Any) -> None:
        pass


class FileLikeStream(Stream[Union[bytes, str]]):
    @abstractmethod
    def open(self,
             *,
             binary_mode: bool = False,
             encoding: Optional[str] = None,
             **kwargs) -> Iterable[Union[bytes, str]]:
        pass


Resource = Union[Stream, Container]



import codecs
import json
from typing import List

from .models import URL

Id = List[str]

MAX_ID_PARTS_COUNT = len(URL._fields)


def to_id(*parts: str,
          max_parts_count: int = MAX_ID_PARTS_COUNT) -> Id:
    if len(parts) > max_parts_count:
        raise ValueError('Parts count should not be '
                         'greater than {max_expected_count}, '
                         'but found {actual_count}.'
                         .format(max_expected_count=max_parts_count,
                                 actual_count=len(parts)))
    return list(parts)


_ENCODING = 'hex'


def serialize_id(id_: Id) -> str:
    return codecs.encode(json.dumps(id_).encode(), _ENCODING).decode()


def deserialize_id(string: str) -> Id:
    return json.loads(codecs.decode(string, _ENCODING).decode())


def to_entry_point(*, id_: Id, module_name: str, function_name: str) -> str:
    return ('{id_} = {module_name}:{function_name}'
            .format(id_=serialize_id(id_),
                    module_name=module_name,
                    function_name=function_name))



from pathlib import Path

from setuptools import (find_packages,
                        setup)

import rsrc

project_base_url = 'https://github.com/lycantropos/rsrc/'

install_requires = Path('requirements.txt').read_text()
setup_requires = [
    'pytest-runner>=4.2',
]
tests_require = Path('requirements-tests.txt').read_text()

setup(name=rsrc.__name__,
      packages=find_packages(exclude=('tests', 'tests.*')),
      version=rsrc.__version__,
      description=rsrc.__doc__,
      long_description=Path('README.md').read_text(encoding='utf-8'),
      long_description_content_type='text/markdown',
      author='Azat Ibrakov',
      author_email='azatibrakov@gmail.com',
      url=project_base_url,
      download_url=project_base_url + 'archive/master.zip',
      python_requires='>=3.5.3',
      install_requires=install_requires,
      setup_requires=setup_requires,
      tests_require=tests_require)
